@@ -1,12 +1,15 @@
 /*
- * api-auth.c
- *
  * Implements the PAM auth group API (pam_sm_authenticate and pam_sm_setcred).
  *
  * The former does and checks the authentication, and the latter creates the
  * final ticket cache and sets its permissions appropriately.  pam_sm_setcred
  * can also refresh an existing ticket cache or destroy a ticket cache,
  * depending on the flags passed in.
+ *
+ * Copyright 2005, 2006, 2007 Russ Allbery <rra@debian.org>
+ * Copyright 2005 Andres Salomon <dilinger@debian.org>
+ * Copyright 1999, 2000 Frank Cusack <fcusack@fcusack.com>
+ * See LICENSE for licensing terms.
  */
 
 /* Get the prototypes for the authentication functions. */
@@ -19,8 +22,13 @@
 #include <krb5.h>
 #include <limits.h>
 #include <pwd.h>
-#include <security/pam_appl.h>
-#include <security/pam_modules.h>
+#ifdef HAVE_SECURITY_PAM_APPL_H
+# include <security/pam_appl.h>
+# include <security/pam_modules.h>
+#elif HAVE_PAM_PAM_APPL_H
+# include <pam/pam_appl.h>
+# include <pam/pam_modules.h>
+#endif
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -182,39 +190,6 @@ done:
     return pamret;
 }
 
-/*
- * If PAM_USER was a fully-qualified principal name, convert it to a local
- * account name and reset it.  This allows users to log in with the full
- * principal that they want to use and let the Kerberos library apply local
- * mapping logic to convert this to an account name.  This should be done
- * after the user is authenticated and authorized.
- *
- * If we fail, don't worry about it, just leave the PAM_USER alone.  It may be
- * that the application doesn't care.
- */
-static void
-canonicalize_name(struct pam_args *args)
-{
-    struct context *ctx = args->ctx;
-    krb5_context c = ctx->context;
-    char kuser[65] = "";        /* MAX_USERNAME == 65 (MIT Kerberos 1.4.1). */
-    char *user;
-    int pamret;
-
-    if (strchr(ctx->name, '@') != NULL) {
-        if (krb5_aname_to_localname(c, ctx->princ, sizeof(kuser), kuser) != 0)
-            return;
-        user = strdup(kuser);
-        if (user == NULL) {
-            pamk5_error(args, "cannot allocate memory: %s", strerror(errno));
-            return;
-        }
-        pamret = pam_set_item(args->pamh, PAM_USER, user);
-        if (pamret != PAM_SUCCESS)
-            pamk5_debug_pam(args, "cannot set PAM_USER", pamret);
-    }
-}
-
 
 /*
  * Authenticate a user via Kerberos 5.
@@ -233,9 +208,9 @@ pam_sm_authenticate(pam_handle_t *pamh, int flags, int argc,
     struct context *ctx = NULL;
     struct pam_args *args;
     krb5_creds *creds = NULL;
-    int pamret;
+    int pamret, ccfd;
     char cache_name[] = "/tmp/krb5cc_pam_XXXXXX";
-    int ccfd;
+    int set_context = 0;
 
     args = pamk5_args_parse(pamh, flags, argc, argv);
     if (args == NULL) {
@@ -249,14 +224,6 @@ pam_sm_authenticate(pam_handle_t *pamh, int flags, int argc,
         goto done;
     ctx = args->ctx;
 
-    /* Do this first so pamk5_context_destroy magically cleans up for us. */
-    pamret = pam_set_data(pamh, "ctx", ctx, pamk5_context_destroy);
-    if (pamret != PAM_SUCCESS) {
-        pamk5_context_free(ctx);
-        pamret = PAM_SERVICE_ERR;
-        goto done;
-    }
-
     /* Check whether we should ignore this user. */
     if (pamk5_should_ignore(args, ctx->name)) {
         pamret = PAM_USER_UNKNOWN;
@@ -275,9 +242,25 @@ pam_sm_authenticate(pam_handle_t *pamh, int flags, int argc,
         goto done;
     }
 
-    /* Store the obtained credentials in a temporary cache. */
+    /* Reset PAM_USER in case we canonicalized. */
+    pamret = pam_set_item(args->pamh, PAM_USER, ctx->name);
+    if (pamret != PAM_SUCCESS)
+        pamk5_debug_pam(args, "cannot set PAM_USER", pamret);
+
+    /* Unless we're creating a ticket cache, we're done. */
     if (args->no_ccache)
         goto done;
+
+    /* Now that we know we're successful, we can store the context. */
+    pamret = pam_set_data(pamh, "pam_krb5", ctx, pamk5_context_destroy);
+    if (pamret != PAM_SUCCESS) {
+        pamk5_context_free(ctx);
+        pamret = PAM_SERVICE_ERR;
+        goto done;
+    }
+    set_context = 1;
+
+    /* Store the obtained credentials in a temporary cache. */
     ccfd = mkstemp(cache_name);
     if (ccfd < 0) {
         pamk5_error(args, "mkstemp(\"%s\") failed: %s", cache_name,
@@ -292,7 +275,6 @@ pam_sm_authenticate(pam_handle_t *pamh, int flags, int argc,
     pamret = set_krb5ccname(args, cache_name, "PAM_KRB5CCNAME");
     if (pamret != PAM_SUCCESS)
         goto done;
-    canonicalize_name(args);
 
 done:
     if (creds != NULL) {
@@ -303,10 +285,16 @@ done:
 
     /*
      * Clear the context on failure so that the account management module
-     * knows that we didn't authenticate with Kerberos.
+     * knows that we didn't authenticate with Kerberos.  Only clear the
+     * context if we set it.  Otherwise, we may be blowing away the context of
+     * a previous successful authentication.
      */
-    if (pamret != PAM_SUCCESS)
-        pam_set_data(pamh, "ctx", NULL, NULL);
+    if (pamret != PAM_SUCCESS) {
+        if (set_context)
+            pam_set_data(pamh, "pam_krb5", NULL, NULL);
+        else
+            pamk5_context_free(ctx);
+    }
     pamk5_args_free(args);
     return pamret;
 }
@@ -388,13 +376,14 @@ static int
 create_session_context(struct pam_args *args)
 {
     struct context *ctx = NULL;
+    PAM_CONST char *user;
     const char *tmpname;
     int status, pamret;
 
     /* If we're going to ignore the user anyway, don't even bother. */
     if (args->ignore_root || args->minimum_uid > 0) {
-        pamret = pam_get_user(args->pamh, &tmpname, NULL);
-        if (pamret == PAM_SUCCESS && pamk5_should_ignore(args, tmpname)) {
+        pamret = pam_get_user(args->pamh, &user, NULL);
+        if (pamret == PAM_SUCCESS && pamk5_should_ignore(args, user)) {
             pamret = PAM_SUCCESS;
             goto fail;
         }
@@ -436,7 +425,7 @@ create_session_context(struct pam_args *args)
      * further calls to session or account management, which OpenSSH does keep
      * the context for.
      */
-    pamret = pam_set_data(args->pamh, "ctx", ctx, pamk5_context_destroy);
+    pamret = pam_set_data(args->pamh, "pam_krb5", ctx, pamk5_context_destroy);
     if (pamret != PAM_SUCCESS) {
         pamk5_debug_pam(args, "cannot set context data", pamret);
         goto fail;
@@ -484,14 +473,16 @@ pam_sm_setcred(pam_handle_t *pamh, int flags, int argc, const char **argv)
      * ticket cache as well.
      */
     if (flags & PAM_DELETE_CRED) {
-        pamret = pam_set_data(pamh, "ctx", NULL, NULL);
+        pamret = pam_set_data(pamh, "pam_krb5", NULL, NULL);
         args->ctx = NULL;
         goto done;
     }
 
     /* If configured not to create a cache, we have nothing to do. */
-    if (args->no_ccache)
+    if (args->no_ccache) {
+        pamret = PAM_SUCCESS;
         goto done;
+    }
 
     /*
      * Reinitialization requested, which means that rather than creating a new
@@ -538,7 +529,7 @@ pam_sm_setcred(pam_handle_t *pamh, int flags, int argc, const char **argv)
      * pam_authenticate, but for either pam_setcred (other than DELETE) or for
      * pam_open_session, the user must be a local account.
      */
-    pw = getpwnam(ctx->name);
+    pw = pamk5_compat_getpwnam(args, ctx->name);
     if (pw == NULL) {
         pamk5_debug(args, "getpwnam failed for %s", ctx->name);
         pamret = PAM_USER_UNKNOWN;
