@@ -1,12 +1,15 @@
 /*
- * auth.c
- *
  * Core authentication routines for pam_krb5.
  *
  * The actual authentication work is done here, either via password or via
  * PKINIT.  The only external interface is pamk5_password_auth, which calls
  * the appropriate internal functions.  This interface is used by both the
  * authentication and the password groups.
+ *
+ * Copyright 2005, 2006, 2007 Russ Allbery <rra@debian.org>
+ * Copyright 2005 Andres Salomon <dilinger@debian.org>
+ * Copyright 1999, 2000 Frank Cusack <fcusack@fcusack.com>
+ * See LICENSE for licensing terms.
  */
 
 #include "config.h"
@@ -14,8 +17,13 @@
 #include <errno.h>
 #include <krb5.h>
 #include <pwd.h>
-#include <security/pam_appl.h>
-#include <security/pam_modules.h>
+#ifdef HAVE_SECURITY_PAM_APPL_H
+# include <security/pam_appl.h>
+# include <security/pam_modules.h>
+#elif HAVE_PAM_PAM_APPL_H
+# include <pam/pam_appl.h>
+# include <pam/pam_modules.h>
+#endif
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -40,32 +48,79 @@
 #include "internal.h"
 
 /*
- * Fill in ctx->princ from the value of ctx->name.
- *
- * This is a separate function rather than just calling krb5_parse_name so
- * that we can work around a bug in MIT Kerberos versions prior to 1.4, which
- * store the realm in a static variable inside the library and don't notice
- * changes.  If no realm is specified and a realm is set in our arguments,
- * append the realm to force krb5_parse_name to do the right thing.
+ * Fill in ctx->princ from the value of ctx->name or (if configured) from
+ * prompting.  If we don't prompt and ctx->name contains an @-sign,
+ * canonicalize it to a local account name.  If the canonicalization fails,
+ * don't worry about it.  It may be that the application doesn't care.
  */
 static krb5_error_code
 parse_name(struct pam_args *args)
 {
     struct context *ctx = args->ctx;
+    krb5_context c = ctx->context;
     char *user = ctx->name;
+    char *newuser;
+    char kuser[65] = "";        /* MAX_USERNAME == 65 (MIT Kerberos 1.4.1). */
     size_t length;
     krb5_error_code k5_errno;
+    int retval;
 
-    if (args->realm != NULL && strchr(ctx->name, '@') == NULL) {
-        length = strlen(ctx->name) + 1 + strlen(args->realm) + 1;
-        user = malloc(length);
-        if (user == NULL)
-            return KRB5_CC_NOMEM;
-        snprintf(user, length, "%s@%s", ctx->name, args->realm);
+    /*
+     * If configured to prompt for the principal, do that first.  Fall back on
+     * using the local username as normal if prompting fails or if the user
+     * just presses Enter.
+     */
+    if (args->prompt_princ) {
+        retval = pamk5_conv(args, "Principal: ", PAM_PROMPT_ECHO_ON, &user);
+        if (retval != PAM_SUCCESS)
+            pamk5_debug_pam(args, "error getting principal", retval);
+        if (*user == '\0') {
+            free(user);
+            user = ctx->name;
+        }
     }
-    k5_errno = krb5_parse_name(ctx->context, user, &ctx->princ);
+
+    /*
+     * We don't just call krb5_parse_name so that we can work around a bug in
+     * MIT Kerberos versions prior to 1.4, which store the realm in a static
+     * variable inside the library and don't notice changes.  If no realm is
+     * specified and a realm is set in our arguments, append the realm to
+     * force krb5_parse_name to do the right thing.
+     */
+    if (args->realm != NULL && strchr(user, '@') == NULL) {
+        length = strlen(user) + 1 + strlen(args->realm) + 1;
+        newuser = malloc(length);
+        if (newuser == NULL)
+            return KRB5_CC_NOMEM;
+        snprintf(newuser, length, "%s@%s", user, args->realm);
+        if (user != ctx->name)
+            free(user);
+        user = newuser;
+    }
+    k5_errno = krb5_parse_name(c, user, &ctx->princ);
     if (user != ctx->name)
         free(user);
+
+    /*
+     * Now that we have a principal to call krb5_aname_to_localname, we can
+     * canonicalize ctx->name to a local name.  We do this even if we were
+     * explicitly prompting for a principal, but we use ctx->name to generate
+     * the local username, not the principal name.  It's unlikely, and would
+     * be rather weird, if the user were to specify a principal name for the
+     * username and then enter a different username at the principal prompt,
+     * but this behavior seems to make the most sense.
+     */
+    if (k5_errno == 0 && strchr(ctx->name, '@') != NULL) {
+        if (krb5_aname_to_localname(c, ctx->princ, sizeof(kuser), kuser) != 0)
+            return 0;
+        user = strdup(kuser);
+        if (user == NULL) {
+            pamk5_error(args, "cannot allocate memory: %s", strerror(errno));
+            return 0;
+        }
+        free(ctx->name);
+        ctx->name = user;
+    }
     return k5_errno;
 }
 
@@ -75,20 +130,41 @@ parse_name(struct pam_args *args)
  * using the Heimdal call to set initial credential options if it's available.
  * This function is used both for regular password authentication and for
  * PKINIT.
+ *
+ * Takes a flag indicating whether we're getting tickets for a specific
+ * service.  If so, we don't try to get forwardable, renewable, or proxiable
+ * tickets.
  */
 static void
-set_credential_options(struct pam_args *args, krb5_get_init_creds_opt *opts)
+set_credential_options(struct pam_args *args, krb5_get_init_creds_opt *opts,
+                       int service)
 {
+#ifdef HAVE_KRB5_MIT
+    /* Work around a bug in MIT Kerberos where allocating the credential
+       structure with opt_alloc doesn't initialize it.  This workaround
+       will probably be removed eventually when the broken versions of 1.6
+       are obsolete.
+
+       We can't do this for Heimdal because it will destroy the private
+       structure in the allocated opt struct. */
+    krb5_get_init_creds_opt_init(opts);
+#endif
 #ifdef HAVE_KRB5_GET_INIT_CREDS_OPT_SET_DEFAULT_FLAGS
     krb5_get_init_creds_opt_set_default_flags(args->ctx->context, "pam",
                                               args->realm_data, opts);
 #endif
-    if (args->forwardable)
-        krb5_get_init_creds_opt_set_forwardable(opts, 1);
-    if (args->lifetime != 0)
-        krb5_get_init_creds_opt_set_tkt_life(opts, args->lifetime);
-    if (args->renew_lifetime != 0)
-        krb5_get_init_creds_opt_set_renew_life(opts, args->renew_lifetime);
+    if (!service) {
+        if (args->forwardable)
+            krb5_get_init_creds_opt_set_forwardable(opts, 1);
+        if (args->lifetime != 0)
+            krb5_get_init_creds_opt_set_tkt_life(opts, args->lifetime);
+        if (args->renew_lifetime != 0)
+            krb5_get_init_creds_opt_set_renew_life(opts, args->renew_lifetime);
+    } else {
+        krb5_get_init_creds_opt_set_forwardable(opts, 0);
+        krb5_get_init_creds_opt_set_proxiable(opts, 0);
+        krb5_get_init_creds_opt_set_renew_life(opts, 0);
+    }
 #ifdef HAVE_KRB5_GET_INIT_CREDS_OPT_SET_PA
     if (args->try_pkinit) {
         if (args->pkinit_user != NULL)
@@ -135,11 +211,11 @@ set_credential_options(struct pam_args *args, krb5_get_init_creds_opt *opts)
  */
 static int
 k5login_password_auth(struct pam_args *args, krb5_creds *creds,
-                      krb5_get_init_creds_opt *opts, char *service,
+                      krb5_get_init_creds_opt *opts, const char *service,
                       char *pass, int *retval)
 {
     struct context *ctx = args->ctx;
-    char *filename;
+    char *filename = NULL;
     char line[BUFSIZ];
     size_t len;
     FILE *k5login;
@@ -148,54 +224,56 @@ k5login_password_auth(struct pam_args *args, krb5_creds *creds,
     int k5_errno;
     krb5_principal princ;
 
-    /* Assume no Kerberos error. */
-    *retval = 0;
-
     /*
      * C sucks at string manipulation.  Generate the filename for the user's
-     * .k5login file.  This function always fails if the user isn't a local
-     * user.
+     * .k5login file.  If the user doesn't exist, the .k5login file doesn't
+     * exist, or the .k5login file cannot be read, fall back on the easy way
+     * and assume ctx->princ is already set properly.
      */
-    pwd = getpwnam(ctx->name);
-    if (pwd == NULL)
-        return PAM_AUTH_ERR;
-    len = strlen(pwd->pw_dir) + strlen("/.k5login");
-    filename = malloc(len + 1);
-    if (filename == NULL)
-        return PAM_SERVICE_ERR;
-    strncpy(filename, pwd->pw_dir, len);
-    filename[len] = '\0';
-    strncat(filename, "/.k5login", len - strlen(pwd->pw_dir));
-
-    /*
-     * If there is no file, do this the easy way.  Assume ctx->princ is
-     * already set properly.
-     */
-    if (access(filename, R_OK) != 0) {
+    pwd = pamk5_compat_getpwnam(args, ctx->name);
+    if (pwd != NULL) {
+        len = strlen(pwd->pw_dir) + strlen("/.k5login");
+        filename = malloc(len + 1);
+    }
+    if (filename != NULL) {
+        strncpy(filename, pwd->pw_dir, len);
+        filename[len] = '\0';
+        strncat(filename, "/.k5login", len - strlen(pwd->pw_dir));
+    }
+    if (pwd == NULL || filename == NULL || access(filename, R_OK) != 0) {
         *retval = krb5_get_init_creds_password(ctx->context, creds,
                      ctx->princ, pass, pamk5_prompter_krb5, args, 0,
-                     service, opts);
+                     (char *) service, opts);
         return (*retval == 0) ? PAM_SUCCESS : PAM_AUTH_ERR;
     }
 
     /*
      * Make sure the ownership on .k5login is okay.  The user must own their
-     * own .k5login or it must be owned by root.
+     * own .k5login or it must be owned by root.  If that fails, set the
+     * Kerberos error code to errno.
      */
     k5login = fopen(filename, "r");
     free(filename);
-    if (k5login == NULL)
+    if (k5login == NULL) {
+        *retval = errno;
         return PAM_AUTH_ERR;
-    if (fstat(fileno(k5login), &st) != 0)
+    }
+    if (fstat(fileno(k5login), &st) != 0) {
+        *retval = errno;
         goto fail;
-    if (st.st_uid != 0 && (st.st_uid != pwd->pw_uid))
+    }
+    if (st.st_uid != 0 && (st.st_uid != pwd->pw_uid)) {
+        *retval = errno;
         goto fail;
+    }
 
     /*
      * Parse the .k5login file and attempt authentication for each principal.
      * Ignore any lines that are too long or that don't parse into a Kerberos
-     * principal.
+     * principal.  Assume an invalid password error if there are no valid
+     * lines in .k5login.
      */
+    *retval = KRB5KRB_AP_ERR_BAD_INTEGRITY;
     while (fgets(line, BUFSIZ, k5login) != NULL) {
         len = strlen(line);
         if (line[len - 1] != '\n') {
@@ -214,7 +292,7 @@ k5login_password_auth(struct pam_args *args, krb5_creds *creds,
         /* Now, attempt to authenticate as that user. */
         *retval = krb5_get_init_creds_password(ctx->context, creds,
                      princ, pass, pamk5_prompter_krb5, args, 0,
-                     service, opts);
+                     (char *) service, opts);
 
         /*
          * If that worked, update ctx->princ and return success.  Otherwise,
@@ -256,7 +334,7 @@ fail:
  * krb5_error_code.  If successful, the credentials will be stored in creds.
  */
 static krb5_error_code
-pkinit_auth(struct pam_args *args, char *service, krb5_creds **creds)
+pkinit_auth(struct pam_args *args, const char *service, krb5_creds **creds)
 {
     struct context *ctx = args->ctx;
     krb5_get_init_creds_opt *opts = NULL;
@@ -295,7 +373,7 @@ pkinit_auth(struct pam_args *args, char *service, krb5_creds **creds)
     retval = krb5_get_init_creds_opt_alloc(ctx->context, &opts);
     if (retval != 0)
         return retval;
-    set_credential_options(args, opts);
+    set_credential_options(args, opts, service != NULL);
 #ifdef HAVE_KRB5_GET_INIT_CREDS_OPT_SET_PKINIT_11_ARGS
     retval = krb5_get_init_creds_opt_set_pkinit(ctx->context, opts,
                   ctx->princ, args->pkinit_user, args->pkinit_anchors, NULL,
@@ -310,7 +388,7 @@ pkinit_auth(struct pam_args *args, char *service, krb5_creds **creds)
 
     /* Finally, do the actual work and return the results. */
     retval = krb5_get_init_creds_password(ctx->context, *creds, ctx->princ,
-                 NULL, pamk5_prompter_krb5, args, 0, service, opts);
+                 NULL, pamk5_prompter_krb5, args, 0, (char *) service, opts);
 
 done:
     pamk5_compat_opt_free(ctx->context, opts);
@@ -324,6 +402,75 @@ done:
 #endif /* HAVE_KRB5_HEIMDAL && HAVE_KRB5_GET_INIT_CREDS_OPT_SET_PKINIT */
 
 
+/*
+ * Try to verify credentials by obtaining and checking a service ticket.  This
+ * is required to verify that no one is spoofing the KDC, but requires read
+ * access to a keytab with a valid key.  By default, the Kerberos library will
+ * silently succeed if no verification keys are available, but the user can
+ * change this by setting verify_ap_req_nofail in [libdefaults] in
+ * /etc/krb5.conf.
+ *
+ * The MIT Kerberos implementation of krb5_verify_init_creds hardwires the
+ * host key for the local system as the desired principal if no principal is
+ * given.  If we have an explicitly configured keytab, instead read that
+ * keytab, find the first principal in that keytab, and use that.
+ *
+ * Returns a Kerberos status code (0 for success).
+ */
+static krb5_error_code
+verify_creds(struct pam_args *args, krb5_creds *creds)
+{
+    krb5_verify_init_creds_opt opts;
+    krb5_keytab keytab = NULL;
+    krb5_kt_cursor cursor;
+    int cursor_valid = 0;
+    krb5_keytab_entry entry;
+    krb5_principal princ = NULL;
+    const char *message;
+    krb5_error_code retval;
+    krb5_context c = args->ctx->context;
+
+    memset(&entry, 0, sizeof(entry));
+    krb5_verify_init_creds_opt_init(&opts);
+    if (args->keytab) {
+        retval = krb5_kt_resolve(c, args->keytab, &keytab);
+        if (retval != 0) {
+            message = pamk5_compat_get_error(c, retval);
+            pamk5_error(args, "cannot open keytab %s: %s", args->keytab,
+                        message);
+            pamk5_compat_free_error(c, message);
+            keytab = NULL;
+        }
+        if (retval == 0)
+            retval = krb5_kt_start_seq_get(c, keytab, &cursor);
+        if (retval == 0) {
+            cursor_valid = 1;
+            retval = krb5_kt_next_entry(c, keytab, &entry, &cursor);
+        }
+        if (retval == 0)
+            retval = krb5_copy_principal(c, entry.principal, &princ);
+        if (retval != 0) {
+            message = pamk5_compat_get_error(c, retval);
+            pamk5_error(args, "error reading keytab %s: %s", args->keytab,
+                        message);
+            pamk5_compat_free_error(c, message);
+        }
+        if (entry.principal != NULL)
+            pamk5_compat_free_keytab_contents(c, &entry);
+        if (cursor_valid)
+            krb5_kt_end_seq_get(c, keytab, &cursor);
+    }
+    retval = krb5_verify_init_creds(c, creds, princ, keytab, NULL, &opts);
+    if (retval != 0)
+        pamk5_error_krb5(args, "credential verification failed", retval);
+    if (princ != NULL)
+        krb5_free_principal(c, princ);
+    if (keytab != NULL)
+        krb5_kt_close(c, keytab);
+    return retval;
+}
+
+
 /*
  * Prompt the user for a password and authenticate the password with the KDC.
  * If correct, fill in creds with the obtained TGT or ticket.  service, if
@@ -333,15 +480,13 @@ done:
  * instead of using PAM_AUTHTOK.
  */
 int
-pamk5_password_auth(struct pam_args *args, char *service, krb5_creds **creds)
+pamk5_password_auth(struct pam_args *args, const char *service,
+                    krb5_creds **creds)
 {
     struct context *ctx;
     krb5_get_init_creds_opt *opts = NULL;
-    krb5_verify_init_creds_opt verify_opts;
-    krb5_keytab keytab = NULL;
     int retval, retry, success;
     char *pass = NULL;
-    const char *message;
     int authtok = (service == NULL) ? PAM_AUTHTOK : PAM_OLDAUTHTOK;
 
     /* Sanity check and initialization. */
@@ -403,7 +548,7 @@ pamk5_password_auth(struct pam_args *args, char *service, krb5_creds **creds)
         pamk5_error_krb5(args, "cannot allocate credential options", retval);
         goto done;
     }
-    set_credential_options(args, opts);
+    set_credential_options(args, opts, service != NULL);
 
     /*
      * If try_first_pass or use_first_pass is set, grab the old password (if
@@ -419,7 +564,7 @@ pamk5_password_auth(struct pam_args *args, char *service, krb5_creds **creds)
     retry = args->try_first_pass ? 1 : 0;
     if (args->try_first_pass || args->use_first_pass || args->use_authtok)
         retval = pam_get_item(args->pamh, authtok, (void *) &pass);
-    if (args->use_authtok && retval != PAM_SUCCESS) {
+    if (args->use_authtok && (retval != PAM_SUCCESS || pass == NULL)) {
         pamk5_debug_pam(args, "no stored password", retval);
         retval = PAM_SERVICE_ERR;
         goto done;
@@ -455,7 +600,7 @@ pamk5_password_auth(struct pam_args *args, char *service, krb5_creds **creds)
         } else {
             retval = krb5_get_init_creds_password(ctx->context, *creds,
                           ctx->princ, pass, pamk5_prompter_krb5, args, 0,
-                          service, opts);
+                          (char *) service, opts);
             success = (retval == 0) ? PAM_SUCCESS : PAM_AUTH_ERR;
         }
         if (success == PAM_SUCCESS) {
@@ -465,38 +610,19 @@ pamk5_password_auth(struct pam_args *args, char *service, krb5_creds **creds)
         }
         pass = NULL;
     } while (retry && retval == KRB5KRB_AP_ERR_BAD_INTEGRITY);
+    if (retval != 0)
+        pamk5_debug_krb5(args, "krb5_get_init_creds_password", retval);
 
 done:
     /*
      * If we think we succeeded, whether through the regular path or via
-     * PKINIT, try to verify the credentials by obtaining and checking a
-     * service ticket.  This is required to verify that no one is spoofing the
-     * KDC, but requires read access to a keytab with an appropriate key.  By
-     * default, the Kerberos library will silently succeed if no verification
-     * keys are available, but the user can change this by setting
-     * verify_ap_req_nofail in [libdefaults] in /etc/krb5.conf.
-     *
-     * Don't do this if we're authenticating for password changes (or any
-     * other case where we're not getting a TGT).  We can't get a service
-     * ticket from a kadmin/changepw ticket.
+     * PKINIT, try to verify the credentials.  Don't do this if we're
+     * authenticating for password changes (or any other case where we're not
+     * getting a TGT).  We can't get a service ticket from a kadmin/changepw
+     * ticket.
      */
-    if (retval == 0 && service == NULL) {
-        krb5_verify_init_creds_opt_init(&verify_opts);
-        if (args->keytab) {
-            retval = krb5_kt_resolve(ctx->context, args->keytab, &keytab);
-            if (retval != 0) {
-                message = pamk5_compat_get_error(ctx->context, retval);
-                pamk5_error(args, "cannot open keytab %s: %s", args->keytab,
-                            message);
-                pamk5_compat_free_error(ctx->context, message);
-                keytab = NULL;
-            }
-        }
-        retval = krb5_verify_init_creds(ctx->context, *creds, NULL, keytab,
-                                        NULL, &verify_opts);
-        if (retval != 0)
-            pamk5_error_krb5(args, "credential verification failed", retval);
-    }
+    if (retval == 0 && service == NULL)
+        retval = verify_creds(args, *creds);
 
     /*
      * If we failed, free any credentials we have sitting around and return
@@ -505,7 +631,6 @@ done:
     if (retval == 0)
         retval = PAM_SUCCESS;
     else {
-        pamk5_debug_krb5(args, "krb5_get_init_creds_password", retval);
         if (*creds != NULL) {
             krb5_free_cred_contents(ctx->context, *creds);
             free(*creds);
@@ -522,6 +647,7 @@ done:
             retval = PAM_ACCT_EXPIRED;
             break;
         case KRB5_KDC_UNREACH:
+        case KRB5_REALM_CANT_RESOLVE:
             retval = PAM_AUTHINFO_UNAVAIL;
             break;
         default:
