@@ -1,11 +1,14 @@
 /*
- * options.c
- *
  * Option handling for pam_krb5.
  *
  * Responsible for initializing the args struct that's passed to nearly all
  * internal functions.  Retrieves configuration information from krb5.conf and
  * parses the PAM configuration.
+ *
+ * Copyright 2005, 2006, 2007 Russ Allbery <rra@debian.org>
+ * Copyright 2005 Andres Salomon <dilinger@debian.org>
+ * Copyright 1999, 2000 Frank Cusack <fcusack@fcusack.com>
+ * See LICENSE for licensing terms.
  */
 
 #include "config.h"
@@ -283,6 +286,7 @@ pamk5_args_parse(pam_handle_t *pamh, int flags, int argc, const char **argv)
         default_string(args, c, "banner", "Kerberos", &args->banner);
         default_string(args, c, "ccache", NULL, &args->ccache);
         default_string(args, c, "ccache_dir", "FILE:/tmp", &args->ccache_dir);
+        default_boolean(args, c, "clear_on_fail", 0, &args->clear_on_fail);
         default_boolean(args, c, "debug", 0, &args->debug);
         default_boolean(args, c, "expose_account", 0, &args->expose_account);
         default_boolean(args, c, "forwardable", 0, &args->forwardable);
@@ -294,6 +298,7 @@ pamk5_args_parse(pam_handle_t *pamh, int flags, int argc, const char **argv)
         default_boolean(args, c, "pkinit_prompt", 0, &args->pkinit_prompt);
         default_string(args, c, "pkinit_user", NULL, &args->pkinit_user);
         default_string(args, c, "preauth_opt", NULL, &preauth_opt);
+        default_boolean(args, c, "prompt_principal", 0, &args->prompt_princ);
         default_time(args, c, "renew_lifetime", 0, &args->renew_lifetime);
         default_boolean(args, c, "retain_after_close", 0, &args->retain);
         default_boolean(args, c, "search_k5login", 0, &args->search_k5login);
@@ -331,6 +336,8 @@ pamk5_args_parse(pam_handle_t *pamh, int flags, int argc, const char **argv)
                 free(args->ccache_dir);
             args->ccache_dir = strdup(&argv[i][strlen("ccache_dir=")]);
         }
+        else if (strcmp(argv[i], "clear_on_fail") == 0)
+            args->clear_on_fail = 1;
         else if (strcmp(argv[i], "debug") == 0)
             args->debug = 1;
         else if (strcmp(argv[i], "expose_account") == 0)
@@ -344,7 +351,7 @@ pamk5_args_parse(pam_handle_t *pamh, int flags, int argc, const char **argv)
         else if (strncmp(argv[i], "keytab=", 7) == 0) {
             if (args->keytab != NULL)
                 free(args->keytab);
-            args->pkinit_anchors = strdup(&argv[i][strlen("keytab=")]);
+            args->keytab = strdup(&argv[i][strlen("keytab=")]);
         }
         else if (strncmp(argv[i], "minimum_uid=", 12) == 0)
             args->minimum_uid = atoi(&argv[i][strlen("minimum_uid=")]);
@@ -371,7 +378,10 @@ pamk5_args_parse(pam_handle_t *pamh, int flags, int argc, const char **argv)
                     = strdup(&argv[i][strlen("preauth_opt")]);
                 args->preauth_opt_count++;
             }
-        } else if (strncmp(argv[i], "realm=", 6) == 0)
+        }
+        else if (strcmp(argv[i], "prompt_principal") == 0)
+            args->prompt_princ = 1;
+        else if (strncmp(argv[i], "realm=", 6) == 0)
             ; /* Handled above. */
         else if (strncmp(argv[i], "renew_lifetime=", 15) == 0) {
             const char *value;
@@ -402,7 +412,6 @@ pamk5_args_parse(pam_handle_t *pamh, int flags, int argc, const char **argv)
         else
             pamk5_error(NULL, "unknown option %s", argv[i]);
     }
-
     if (flags & PAM_SILENT)
         args->silent = 1;
 
@@ -421,6 +430,10 @@ pamk5_args_parse(pam_handle_t *pamh, int flags, int argc, const char **argv)
     if (args->search_k5login)
         args->expose_account = 0;
 
+    /* UIDs are unsigned on some systems. */
+    if (args->minimum_uid < 0)
+        args->minimum_uid = 0;
+
     /*
      * Warn if PKINIT options were set and PKINIT isn't supported.  The MIT
      * method (krb5_get_init_creds_opt_set_pa) can't support use_pkinit.
